from matplotlib import pyplot as plt
from scipy.integrate import solve_ivp
from scipy.optimize import fsolve
import numpy as np
import pandas as pd
import hssm

hssm.set_floatX("float32")

def run_ddm(drift_rate = 1,
            threshold = 1,
            starting_point = 0.5,
            noise_std = 1,
            non_decision_time = 0,
            dt = 0.01,
            max_time = 200):
    """
    Simulates the Drift Diffusion Model for one run with fixed time increments to match evidence points.

    Parameters are the same as the previous function.

    Arguments:
    - drift_rate: rate of evidence accumulation
    - threshold: evidence needed to reach a decision
    - starting_point: initial condition of the evidence
    - noise_std: standard deviation of the noise term
    - non_decision_time: time not included in the evidence accumulation process
    - dt: time increment
    - max_time: maximum simulation time

    Returns:
    - decision_time: the time taken to reach a decision
    - evidence_over_time: the evidence accumulated over time
    - decision: 1 if the decision boundary reached is the upper threshold, -1 if lower
    """
    # Initialize evidence accumulation process
    time = non_decision_time
    evidence = starting_point
    evidence_over_time = [evidence]

    # Run the simulation until a decision boundary is reached or max time is exceeded)
    while time < max_time:
        # Increment the evidence by the drift and some noise
        evidence += drift_rate * dt + np.sqrt(dt) * noise_std * np.random.normal() # THERE IS A BUG IN THIS LINE
        # evidence += drift_rate + dt + np.sqrt(dt) * noise_std * np.random.normal()
        evidence_over_time.append(evidence)
        time += dt  # Increment the time

        # Check if the evidence has reached either threshold
        if evidence >= threshold: # THERE IS A BUG IN THIS LINE
            return time, evidence_over_time, 1  # Decision made for the upper threshold
        elif evidence <= 0:
            return time, evidence_over_time, -1  # Decision made for the lower threshold

    # If no decision has been made by max_time, return the current state
    return time, evidence_over_time, 0

def test_run_ddm(run_ddm_student):

    v = 1
    a = 1
    z = 0.5
    sigma = 1
    t0 = 0
    dt = 0.01
    max_time = 200

    np.random.seed(0)
    decision_time, evidence_over_time, decision = run_ddm(drift_rate=v,
                                                          threshold=a,
                                                          starting_point=z,
                                                          noise_std=sigma,
                                                          non_decision_time=t0,
                                                          dt=dt,
                                                          max_time=max_time)

    np.random.seed(0)
    student_decision_time, student_evidence_over_time, student_decision = run_ddm_student(drift_rate=v,
                                                                                         threshold=a,
                                                                                         starting_point=z,
                                                                                         noise_std=sigma,
                                                                                         non_decision_time=t0,
                                                                                         dt=dt,
                                                                                         max_time=max_time)

    if not np.allclose(decision_time, student_decision_time, atol=0.01):
        print("Your run_ddm function generates incorrect outputs. Check for mistakes.")
    else:
        print("Your run_ddm function generates correct outputs!")

def get_simulated_histogram_data():
    v_true, a_true, z_true, t_true = [-0.5, 1, 0.5, 0.5]
    dataset = hssm.simulate_data(
        model="ddm",
        theta=[v_true, a_true, z_true, t_true],
        size=1000,
    )

    return dataset

def test_simulated_ddm(v, a, t0, rt_upper, rt_lower):
    looks_good = True

    if v >= 0:
        looks_good = False

    if t0 < 0.3 or t0 > 0.7:
        looks_good = False

    if looks_good:
        print("Your parameters look reasonable.")
    else:
        print("Your parameters may be wrong.")


def get_dot_data_set():
    # Set up trial by trial parameters
    intercept = 0.2
    x = np.random.uniform(-1, 1, size=500)
    v = intercept + (0.5 * x)

    true_values = np.column_stack(
        [v, np.repeat([[1.5, 0.5, 0.5, 0.0]], axis=0, repeats=500)]
    )

    dataset = hssm.simulate_data(
        model="ddm",
        theta=true_values,
        size=1,
    )

    # dataset
    dataset_reg_v = pd.DataFrame(
        {
            "rt": dataset["rt"],
            "response": dataset["response"],
            "coherence": x,
        }
    )

    return dataset_reg_v

def get_threshold_model(coherence_dataset):
    name = "a"
    formula = "a ~ 1 + coherence"

    varying_threshold_model = hssm.HSSM(
        data=coherence_dataset,
        include=[
            {
                "name": name,
                "formula": formula,
            }
        ],
    )
    return varying_threshold_model

def test_threshold_model(name, formula):
    looks_good = True

    if name != "a":
        looks_good = False

    # remove spaces from formula
    formula = formula.replace(" ", "")

    if formula != "a~1+coherence":
        looks_good = False

    if looks_good:
        print("Your model specification is correct.")
    else:
        print("Your model specification is wrong. Check for mistakes.")

def S(g,y):
  """Logistic saturation function, with slope modulated by a gain parameter

  Inputs
  --------
  g: positive scalar neuronal gain
  y: net input to the neural processing unit
  """

  return 1/(1 + np.exp(-g*y))

def sim_trajectories(t, x, g, I1, I2, sigma):
  """ Dynamical model

  ------
  t: scalar time (you can ignore this variable)
  x: array [x1,x2] of activities of the neural units corresponding to tasks 1,2
  w_{12} = w{21} = 1
  g: gain
  I1: input to x_1 (cue for task 1)
  I2: input to x_2 (cue for task 2)
  sigma: standard deviation of Gaussian noise added to the net input


  Output
  -------
  dx1/dt(t,x): - x1 + S(-w_{12} x2 + I1)
  dx2/dt(t,x): - x2 + S(-w_{21} x1 + I2)    where S(y) = 1/(1 + e^(-g y))

  """
  # inhibitory weights
  w1_2=1
  w2_1=1

  # initial conditions
  x1 = x[0]
  x2 = x[1]

  ###Exercise 1.c START###

  dx1_dt = - x1 + S(g, -w1_2 * x2 + I1 + sigma * np.random.normal()) # YOUR CODE GOES HERE   (0.5 points)

  dx2_dt = - x2 + S(g, -w2_1 * x1 + I2 + sigma * np.random.normal()) # YOUR CODE GOES HERE   (0.5 points)

  ###Exercise 1.c END###

  return np.array([dx1_dt, dx2_dt])

def test_sim_trajectories(sim_trajectories_student):

    t = 0
    np.random.seed(0)
    x_0 = np.array([np.random.uniform(),np.random.uniform()])        # Initial condition (x1(0), x2(0)) (must have the same number of entries as your list of variables!)
    I1 = 1.2; I2 = 0.5; g = 5          # parameters
    sigma = 0.1

    np.random.seed(0)
    dx1_dt, dx2_dt = sim_trajectories(t, x_0, g, I1, I2, sigma)

    np.random.seed(0)
    dx1_dt_student, dx2_dt_student = sim_trajectories_student(t, x_0, g, I1, I2, sigma)

    if not np.allclose(dx1_dt, dx1_dt_student, atol=0.01):
        print("Your sim_trajectories function generates incorrect outputs. Check for mistakes.")
    elif not np.allclose(dx2_dt, dx2_dt_student, atol=0.01):
        print("Your sim_trajectories function generates incorrect outputs. Check for mistakes.")
    else:
        print("Your sim_trajectories function generates correct outputs!")

def compute_dx_dt(x_i, x_j, g, I, sigma):
  """ Activation Difference Function

  ------
  x_i: activity of the neural unit x_i corresponding to task 1 or task 2
  x_j: activity of the other neural unit x_i corresponding to task 2 or task 1
  g: gain
  I: input to the neural unit x_i
  sigma: standard deviation of Gaussian noise added to the net input

  Output
  -------
  dxi/dt
  """

  dx_dt = - x_i + S(g, -1 * x_j + I + sigma * np.random.normal())

  return dx_dt

def plot_trajectory(T, g, I1, I2, sigma, x1_0, x1_1, ax=None):
  # Initial condition (x1(0), x2(0)) (must have the same number of entries as your list of variables!)
  x_0 = np.array([x1_0, x1_1])

  # Integrate differential equation
  x_out = solve_ivp(sim_trajectories,
                    np.array([0,T]),
                    x_0,                # initial condition
                    dense_output=True,  # dense_output = True allows us to compute x at any time points on the interval T
                    args = [g, I1, I2, sigma]) # pass additional arguments to the simulation functon

  # First, extract values of the simulation at the desired time grid:
  ts = np.linspace(0,T,100)         # list of 100 evenly spaced points in the time interval we are considering
  xt = x_out.sol(ts)                # solution of the integral at the specified time points
  x1 = xt[0,:]                      # the values of x1(t) are in the first column of the matrix xt
  x2 = xt[1,:]                      # the values of x2(t) are in the second column of the matrix xt

  if ax is None:
    # Next, visualize the trajectories:
    plt.plot(ts,x1,label='x1')        # this plots the time trajectory of x1
    plt.plot(ts,x2,label='x2')        # this plots the time trajectory of x2
    plt.legend()                      # show the legend
    plt.xlabel('t')                   # this labels the horizontal axis
    plt.ylim([0, 1.0])                # this sets the upper and lower bounds for the vertical axis
  else:
    ax.plot(ts,x1,label='x1')        # this plots the time trajectory of x1
    ax.plot(ts,x2,label='x2')        # this plots the time trajectory of x2
    ax.legend()                      # show the legend
    ax.set_xlabel('t')                   # this labels the horizontal axis
    ax.set_ylim([0, 1.0])                # this sets the upper and lower bounds for the vertical axis

def plot_control_participant():
    # Parameters
    T = 50  # Length of time for the simulation (keep this fixed for now)
    g = 10  # Gain parameter (you may tweak this for the exercise)
    I1 = 0  # Input to task 1 (keep this fixed for now)
    I2 = 0  # Input to task 2 (keep this fixed for now)
    x1_0 = 1  # Initial condition of task 1 variable (you may tweak this for the exercise)
    x1_1 = 0  # Initial condition of task 1 variable (you may tweak this for the exercise)

    np.random.seed(0)

    # plot noise-less environment
    sigma = 0
    fig, axs = plt.subplots(1, 2, figsize=(10, 5))  # 1 row, 2 columns
    plot_trajectory(T, g, I1, I2, sigma, x1_0, x1_1, axs[0])
    axs[0].set_title('Healthy Patient Without Noise')

    # plot noisy environment
    sigma = 0.2
    plot_trajectory(T, g, I1, I2, sigma, x1_0, x1_1, axs[1])
    axs[1].set_title('Healthy Patient With Noise')

    plt.tight_layout()
    plt.show()


import numpy as np # for doing math easily in python
import matplotlib.pyplot as plt # for plotting in python
import itertools


def bayesRule(probHypothesis, probObservation):
  '''
 Calculate the posterior probability with Bayes' Rule.
 Arguments:
  probHypothesis (float): A list of the prior distribution.
  probObservation (float): The probability of the observation.
 Returns:
  probPosterior (float): The probability value of the posterior.
  '''

  jointProb = 0.025
  likelihood = jointProb / probObservation
  probPosterior = likelihood * probHypothesis / probObservation


  return probPosterior

def probabilityDensityFunction(mean, variance, x):
  '''
  Calculates the probability density function in Gaussian form.
  The function returns the inputs mean and variance for convenience to be used later.

  Arguments:
    x(np.array): The input array of the function.
    mean (float): The mean of the function.
    variance (float): The variance of the function, sigma squared.

  Returns:
    mean (float): The given mean of the function.
    varinace (float): The given variance of the function.
    pdf (np.array): The output of the function as the probability density function.

  '''

  pdf = 1 / np.sqrt(2*np.pi*variance) * np.exp(-(x-mean)**2/(2*variance)) # YOUR CODE GOES HERE
  return pdf

def test_probabilityDensityFunction(student_function):
    mean = 0
    variance = 1
    x = np.linspace(-1,1,100)
    pdf_student = student_function(mean, variance, x)
    pdf = probabilityDensityFunction(mean, variance, x)
    isclose = np.allclose(pdf_student, pdf)

    if not isclose:
        print("Your function is not computing correct outputs. Check for mistakes.")
        return
    print("Your function is computing correct outputs!")
    return

def multiplyDistributions(mean1, mean2, var1, var2, x):
  '''
  Multiply two normal distributions.
  Arguments:
    x (np.array): The x values of the distribution.
    mean1 (float): The mean of the first distribution.
    mean2 (float): The mean of the second distribution.
    var1 (float): The variance of the first distribution, sigma squared.
    var2 (float): The variance of the second distribution, sigma sqaured.

  Returns:
    mean_post (float): The mean of the resulting distribution.
    var_post(float): The variance of the resulting distribution.
  '''


  # Define the probability density functions
  mean1, var1, pdf1 = probabilityDensityFunction(mean1, var1, x)
  mean2, var2, pdf2 = probabilityDensityFunction(mean2, var2, x)

  # Calculate the mean and the variance of the probability density function for the posterior
  meanPosterior = (1/var1 * mean1 + 1/var2 * mean2) / (1/var1 + 1/var2)
  varPosterior = 1 / (1/var1 + 1/var2)

  return meanPosterior, varPosterior

def test_multiplyDistributions(student_function):
    student_function = multiplyDistributions
    mean1 = 0
    mean2 = 0
    var1 = 1
    var2 = 1
    x = np.linspace(-1,1,100)
    mean_post_student, var_post_student = student_function(mean1, mean2, var1, var2, x)
    mean_post, var_post = multiplyDistributions(mean1, mean2, var1, var2, x)
    isclose = np.allclose(mean_post_student, mean_post) and np.allclose(var_post_student, var_post)

    if not isclose:
        print("Your function is not computing correct outputs. Check for mistakes.")
        return
    print("Your function is computing correct outputs!")
    return


def bayesUpdate(meanPrior, meanLikelihood, varPrior, varLikelihood, x):

  pdfPrior = probabilityDensityFunction(meanPrior, varPrior, x)
  meanPosterior, varPosterior = multiplyDistributions(meanPrior, meanLikelihood, varPrior, varLikelihood, x) # YOUR CODE GOES HERE

  # Bayesian update

  meanPrior = meanPosterior  # YOUR CODE GOES HERE
  varPrior = varPosterior   # YOUR CODE GOES HERE

  return meanPrior, varPrior, pdfPrior

def test_bayesUpdate(student_function):
    meanPrior = 0
    meanLikelihood = 0
    varPrior = 1
    varLikelihood = 1
    x = np.linspace(-1,1,100)
    mean_post_student, var_post_student, pdf_student = student_function(meanPrior, meanLikelihood, varPrior, varLikelihood, x)
    mean_post, var_post, pdf = bayesUpdate(meanPrior, meanLikelihood, varPrior, varLikelihood, x)

    isclose = np.allclose(mean_post_student, mean_post) and np.allclose(var_post_student, var_post) and np.allclose(pdf_student, pdf)

    if not isclose:
        print("Your function is not computing correct outputs. Check for mistakes.")
        return
    print("Your function is computing correct outputs!")
    return

def bipartite_to_dependency(adjacency_matrix_of_bipartite_graph):

  A = adjacency_matrix_of_bipartite_graph

  # determine number of tasks
  num_tasks = int(np.sum(A.flatten()))

  # for each of the tasks below, let's us store their input and output dimensions
  # these will come in handy later on
  task_list = list()
  for row in range(A.shape[0]):
    for col in range(A.shape[1]):
      if A[row, col] == 1: # we found a task
        # we are appending each task as a tuple specifying its input and output dimension
        task_list.append((row, col))

  # set up dependency graph
  dependency_graph = np.eye(num_tasks)

  # we will now enumerate through each pair of tasks
  for task1_id, input_output_tuple1 in enumerate(task_list):
    for task2_id, input_output_tuple2 in enumerate(task_list):
      if task1_id == task2_id:
        continue

      dependency = 0 # let's begin with assuming no dependency

      # first check if the two tasks are structurally dependent,
      # i.e., whether they share an input or output dimension
      if input_output_tuple1[0] == input_output_tuple2[0] or input_output_tuple1[1] == input_output_tuple2[1]:
        dependency = 1

      if dependency == 0:
        # if the task is not structurally dependent, then it may be functionally dependent
        # to determine functional dependence, we have to check if there is a third task in A
        # that either shares an input dimension our an output dimension with our task
        for task3_id, input_output_tuple3 in enumerate(task_list):
          if (input_output_tuple3[0] == input_output_tuple1[0] and input_output_tuple3[1] == input_output_tuple2[1]) or (input_output_tuple3[0] == input_output_tuple2[0] and input_output_tuple3[1] == input_output_tuple1[1]):
            dependency = 1
            break

      dependency_graph[task1_id, task2_id] = dependency
      dependency_graph[task2_id, task1_id] = dependency

  return dependency_graph

def test_bipartite_graph(bipartite_graph_student):
    bipartite_graph = np.array([[1, 0, 0],
                                [0, 1, 0],
                                [1, 0, 1],
                                [0, 1, 0]])

    equal = np.array_equal(bipartite_graph, bipartite_graph_student)

    if equal is True:
        print("Your adjancency matrix is correct!")
    else:
        print("Your adjancency matrix is incorrect. Check for mistakes.")

def get_example_dependency_graph():
    bipartite_graph = np.array([[1, 0, 0],
                                [0, 1, 0],
                                [1, 0, 1],
                                [0, 1, 0]])

    return bipartite_to_dependency(bipartite_graph)

def test_multitaskable_tasks(multitaskable_tasks):
    multitaskable_tasks = list(dict.fromkeys(multitaskable_tasks))

    dependency_graph = get_example_dependency_graph()
    edges = ['a', 'b', 'c', 'd', 'e']
    multitaskable = list()

    for row in range(dependency_graph.shape[0]):
        for col in range(row, dependency_graph.shape[0]):

            if dependency_graph[row, col] == 0:
                multitaskable.append((edges[row], edges[col]))

    for proposed_task_pair in multitaskable_tasks:

        found = False
        for idx, task_pair in enumerate(multitaskable):
            print(task_pair)
            if (proposed_task_pair[0] == task_pair[0] and proposed_task_pair[1] == task_pair[1]) or (
                    proposed_task_pair[0] == task_pair[1] and proposed_task_pair[1] == task_pair[0]):
                found = True
                del multitaskable[idx]
                break

        if found is False:
            print("Your list of multitaskable tasks is incomplete or incorrect. Check for mistakes.")
            return

    if len(multitaskable) == 0:
        print("Your list of multitaskable tasks is correct!")
    else:
        print("Your list of multitaskable tasks is incomplete or incorrect. Check for mistakes.")

def add_shared_representations(adjacency_matrix, p):
    for row in range(adjacency_matrix.shape[0]):
        for col in range(adjacency_matrix.shape[1]):
            if row == col:
                continue
            else:
                randnum = np.random.rand()
                if randnum < p:
                    adjacency_matrix[row, col] = 1
                    adjacency_matrix[col, row] = 1
    return adjacency_matrix

def is_independent_set(adjacency_matrix, node_set):
    """Check if a set of nodes is an independent set."""
    for i in node_set:
        for j in node_set:
            if i != j and adjacency_matrix[i, j] == 1:
                return False
    return True

def maximum_independent_set(adjacency_matrix, verbose=False):
    """Find the maximum independent set of a graph given its adjacency matrix."""
    n = len(adjacency_matrix)
    max_set = set()

    for size in range(1, n + 1):
        for subset in itertools.combinations(range(n), size):
            if verbose:
                print(str(subset) + str(is_independent_set(adjacency_matrix, subset)))
            if is_independent_set(adjacency_matrix, subset):
                if len(subset) > len(max_set):
                    max_set = set(subset)
    return max_set

def get_MIS_cardinality(adjacency_matrix):
    """determine the cardinality of the maximum independent set"""
    max_set = maximum_independent_set(adjacency_matrix)

    return len(max_set)

def compute_parallel_processing_capability(N, p):
    '''
    This function computes the parallel procesing capability of a graph.

    Arguments:
    - N: number of input/output nodes in the bipartite graph
    - p: edge probability (index for representation sharing)

    Returns:
    - MIS_cardinality: cardinality of the maximum independent set (single number)
    '''

    bipartite_graph = np.eye(N)
    bipartite_graph = add_shared_representations(bipartite_graph, p)
    dependency_graph = bipartite_to_dependency(bipartite_graph)
    MIS_size = get_MIS_cardinality(dependency_graph)

    return MIS_size

def test_compute_parallel_processing_capability(student_function):

    passed = True

    N = 5
    p = 0.5
    np.random.seed(0)
    MIS_size = compute_parallel_processing_capability(N, p)
    np.random.seed(0)
    student_MIS_size = student_function(N, p)
    if MIS_size != student_MIS_size:
        passed = False

    N = 4
    p = 0
    np.random.seed(0)
    MIS_size = compute_parallel_processing_capability(N, p)
    np.random.seed(0)
    student_MIS_size = student_function(N, p)
    if MIS_size != student_MIS_size:
        passed = False

    N = 4
    p = 1
    np.random.seed(0)
    MIS_size = compute_parallel_processing_capability(N, p)
    np.random.seed(0)
    student_MIS_size = student_function(N, p)
    if MIS_size != student_MIS_size:
        passed = False

    if passed is True:
        print("It looks like your function is computing correct outputs.")
    else:
        print("Your function is not computing correct outputs. Check for mistakes.")


def compute_E(correct_response_probability, N):
  num_keys = N
  p_resp = correct_response_probability
  p_nonresp = (1-p_resp)/(num_keys-1)
  p_prior = 1/num_keys
  return (num_keys-1) * p_nonresp * np.log(p_nonresp/p_prior) + p_resp * np.log(p_resp/p_prior)

def test_compute_E(student_function):
  for p in np.linspace(0.01, 0.99, 10):
    for N in range(2, 10):
      if not np.isclose(compute_E(p, N), student_function(p, N)):
        print("Your function is not computing correct outputs. Check for mistakes.")
        return
  print("It looks like your function is computing correct outputs.")
  return


"""
authors: Sebastian Musslick (smusslick@uos.de),
         Deniz M. Gun (dguen@uos.de)

This module contains helper and test functions for Problemset 1
 of the course "Cognitive Modeling" at the University of Osnabrueck.
 This module exists in order to load certain functionality into the 
 assignment notebooks without involuntarily giving students access to
 solutions.
"""
import numpy as np
import random
import pandas as pd


def compute_valence(deck_result, w):
    '''
    This function computes the valence as denoted in the notebook script.

    Input arguments:
      deck_result: The result of the selected deck (in points)
      W: The valence parameter
    '''

    # first compute win and loss depending on points received
    if deck_result > 0:
        win = deck_result
        loss = 0
    else:
        win = 0
        loss = deck_result

    # compute valence term
    valence = w * win + (1-w) * loss  # YOUR ANSWER GOES HERE

    return valence


def test_compute_valence(student_valence, print_feedback=True):
    """ Returns True if The student implemented 
    valence function returns same values as the correct solution.

    Args
        student_valence (method) :  
            The compute_valence method defined in the notebook and edited by students.
        print_feedback (bool) : 
            If True, this method prints feedback to the console.
    Returns
        correct (bool): 
            True, if student_valence returns the same values as compute_valence
    """

    correct = True

    # Search for all reck rewards between -25 and 10
    for deck_result in range(-25, 10, 1):
        for w in range(11):
            w = w/10  # scaling w to [0,1]
            student_result = student_valence(deck_result, w)
            correct_result = compute_valence(deck_result, w)

            if student_result != correct_result:
                correct = False
                break

        if not correct:
            break  # exit search

    if not print_feedback:
        return correct

    if correct:
        print("Your compute_valence function creates correct outputs!")
    else:
        print(
            "Your compute_valence function generates incorrect outputs. Check for mistakes.")
    return correct


def compute_expectancy(previous_expectancy, alpha, valence):
    '''
    This function computes the expectancy as denoted above.

    Input arguments:
      previous_expectancy: expectancy of current deck on previous trial
      alpha: update parameter
      valence: valence of current deck on current trial
    '''

    # compute expectancy term
    new_expectancy = (1-alpha) * previous_expectancy + \
        alpha * valence  # YOUR ANSWER GOES HERE
    return new_expectancy


def test_compute_expectancy(student_expectancy, print_feedback=True):
    """ Returns True if The student implemented
    expectancy function returns same values as the correct solution.

    Args
        student_expectancy (method) :
            The compute_expectancy method defined in the notebook and edited by students.
        print_feedback (bool) :
            If True, this method prints feedback to the console.
    Returns
        correct (bool):
            True, if student_expectancy returns the same values as compute_expectancy
    """

    correct = True

    for previous_expectancy in range(-10, 10, 1):
        for alpha in range(11):
            alpha = alpha / 10  # scaling to [0,1]
            for valence in range(11):
                valence = valence/10  # scaling to [0,1]
                student_result = student_expectancy(
                    previous_expectancy, alpha, valence)
                correct_result = compute_expectancy(
                    previous_expectancy, alpha, valence)

                if student_result != correct_result:
                    correct = False
                    break

        if not correct:
            break  # exit search

    if not print_feedback:
        return correct

    if correct:
        print("Your compute_expectancy function creates correct outputs!")
    else:
        print(
            "Your compute_expectancy function generates incorrect outputs. Check for mistakes.")
    return correct


def compute_choice_probabilities(expectancies, trial_number, c):
    '''
    This function computes the choice probabilities for each deck based on the
    current expectancies

    Input arguments:
      expectancies: a list of expectancies for all four decks
      trial_number: number of the current trial
      c: parameter used to compute beta
    '''
    # compute beta
    beta = pow((trial_number+1)/10, c)

    # compute softmaxed choice proabbilities
    # YOUR ANSWER GOES HERE
    choice_probs = np.exp(expectancies * beta) / \
        np.sum(np.exp(expectancies * beta))

    return choice_probs


def test_compute_choice_probs(student_choice_probabilities, print_feedback=True):
    """ Returns True if The student implemented
    choice probability function returns same values as the correct solution.

    Args
        student_choice_probabilities (method) :
            The compute_choice_probabilities method defined in the notebook and edited by students.
        print_feedback (bool) :
            If True, this method prints feedback to the console.
    Returns
        correct (bool):
            True, if student_choice_probabilities returns the same values as compute_choice_probabilities
    """

    correct = True

    for expectancy_idx in range(11):
        for trial_number in range(11):
            for c in np.linspace(-2, 2, 11):

                expectancies = np.array([np.random.randint(-5, 5),
                                        np.random.randint(-5, 5),
                                        np.random.randint(-5, 5),
                                         np.random.randint(-5, 5)])
                student_result = student_choice_probabilities(
                    expectancies, trial_number, c)
                correct_result = compute_choice_probabilities(
                    expectancies, trial_number, c)

                for choice in range(len(student_result)):
                    if np.round(student_result[choice], decimals=8) != np.round(correct_result[choice], decimals=8):
                        correct = False
                        break

        if not correct:
            break  # exit search

    if not print_feedback:
        return correct

    if correct:
        print("Your compute_choice_probabilities function creates correct outputs!")
    else:
        print(
            "Your compute_choice_probabilities function generates incorrect outputs. Check for mistakes.")
    return correct


def compute_choice(expectancies, trial_number, c):
    '''
    This function computes the index of the chosen deck for the next trial

    Input arguments:
      expectancies: a list of expectancies for all four decks
      trial_number: number of the current trial
      c: parameter used to compute beta
    '''

    # compute softmaxed choice proabbilities
    choice_probs = compute_choice_probabilities(expectancies, trial_number, c)

    # compute cumulated sums
    cumulated_choice_probs = np.cumsum(choice_probs)

    # draw random number between 0 and 1
    random_number = random.random()

    # choose deck index depending on choice probabilities
    index = 0

    # Iterate through the cumulative sums to find the first index where the random number exceeds the cumulative sum
    while index < len(cumulated_choice_probs) and random_number > cumulated_choice_probs[index]:
        index += 1

    return index


def observe_deck_result(wins, losses, win_probability, choice):
    '''
    This function returns the outcome of choosing a deck.

    Input arguments:
      wins: wins for Decks A, B, C, and D
      losses: losses for Decks A, B, C, and D
      win_probability: chance of obtaining a win from a deck
      choice: the index of the chosen deck (0: A, 1: B, 2: C, 3: D)
    '''
    random_number = random.random()
    if random_number <= win_probability:
        return wins[choice]
    else:
        return losses[choice]


def log_data(df, trial_number, expectancies, choice, wins, losses, deck_result):
    '''
    This function logs several outcomes of the simulation into a data frame.

    Input arguments:
      df: a Pandas dataframe in which we write the data
      trial_index: current trial number
      expectancies: expectancies of all decks
      choice: the index of the chosen deck (0: A, 1: B, 2: C, 3: D)
      wins: wins for Decks A, B, C, and D
      losses: losses for Decks A, B, C, and D
      deck_result: the value obtained from choosing the deck
    '''
    if wins[choice] == np.min(wins):
        type = "safe"
    else:
        type = "risky"

    new_row = pd.DataFrame({'trial_index': trial_number,
                            'E(A)': expectancies[0],
                            'E(B)': expectancies[1],
                            'E(C)': expectancies[2],
                            'E(D)': expectancies[3],
                            'choice_index': choice,
                            'type': type,
                            'reward': wins[choice],
                            'penalty': losses[choice],
                            'value': deck_result}, index=[0])

    df = pd.concat([df, new_row], axis=0, ignore_index=True)
    return df


def simulate_model(num_trials, wins, losses, win_probability, W, alpha, c):
    '''
    This function simulates the expectancy valence model for a certain number of trials.
    It returns a dataframe with simulation results.

    Input arguments:
      wins: wins for Decks A, B, C, and D
      losses: losses for Decks A, B, C, and D
      win_probability: chance of obtaining a win from a deck
      W: the valence parameter
      alpha: the update parameter
      c: parameter used to compute the softmax parameter beta
    '''
    # initialize the model
    expectancies = np.zeros(4)

    # we will log the entire simulation in a dataframe
    df = pd.DataFrame(columns=['trial_index', 'E(A)', 'E(B)',
                      'E(C)', 'E(D)', 'choice_index', 'type', 'value'])

    for trial_number in range(num_trials):

        # make a choice
        chosen_deck_index = compute_choice(expectancies, trial_number, c)

        # REPLACE: problemset1.compute_choice
        # YOUR ANSWER GOES HERE
        # Hint: You may use the functions observe_deck_result, compute_valence, and compute_expectancy
        # If your compute_valence and compute_expectancy don't pass the tests above, then you may use
        # the problemset1.compute_valence and problemset1.compute_expectancy functions instead (they have
        # the same inputs and outputs).

        # observe an outcome
        deck_result = observe_deck_result(
            wins, losses, win_probability, chosen_deck_index)

        # compute valence
        valence = compute_valence(deck_result, W)

        # update expectancy of chosen deck
        expectancies[chosen_deck_index] = compute_expectancy(
            expectancies[chosen_deck_index], alpha, valence)

        # log results
        df = log_data(df, trial_number, expectancies,
                      chosen_deck_index, wins, losses, deck_result)

    return df


def test_simulate_model(student_simulate_model, print_feedback=True):
    """ Returns True if The student implemented
    simulate model function returns same values as the correct solution.

    Args
        simulate_model (method) :
            The simulate_models method defined in the notebook and edited by students.
        print_feedback (bool) :
            If True, this method prints feedback to the console.
    Returns
        correct (bool):
            True, if simulate_model returns the same values as simulate_model
    """

    correct = True

    wins = [5, 10, 10, 5]  # wins for Decks A, B, C, and D
    losses = [-5, -25, -25, -5]  # losses for Decks A, B, C, and D
    win_probability = 0.5  # chance of obtaining a win from a deck
    num_trials = 20

    for W in np.linspace(0, 1, 4):
        for alpha in np.linspace(0, 1, 4):
            for c in np.linspace(-1, 1, 4):

                random.seed(1)
                student_result = student_simulate_model(
                    num_trials, wins, losses, win_probability, W, alpha, c)
                random.seed(1)
                correct_result = simulate_model(
                    num_trials, wins, losses, win_probability, W, alpha, c)

                if student_result.equals(correct_result) is False:
                    correct = False
                    break

        if not correct:
            break  # exit search

    if not print_feedback:
        return correct

    if correct:
        print("Your simulate_model function creates correct outputs!")
    else:
        print(
            "Your simulate_model function generates incorrect outputs. Check for mistakes.")
    return correct


def compute_log_likelihood(df, W, alpha, c):
    '''
    This function computes the log likelihood for a given set of model parameters

    Input arguments:
      df: data frame
      W: the valence parameter
      alpha: the update parameter
      c: parameter used to compute the softmax parameter beta
    '''
    # initialize log likelihood
    LL = 0

    # initialize the model
    expectancies = np.zeros(4)

    num_trials = len(df)

    for trial_number in range(num_trials):

        # when computing the likelihood, we don't need to make a choice
        # instead, we observe the choice from the data
        chosen_deck_index = df['choice_index'][trial_number]

        # next, we compute the log likelihood of that choice given the model and its parameters
        # the log likelihood is simply the log of the probability of choosing the current deck

        choice_probabilities = compute_choice_probabilities(
            expectancies, trial_number, c)

        # YOUR ANSWER GOES HERE
        LL = LL + np.log(choice_probabilities[chosen_deck_index])

        # next, we observe the result obtained from that trial and compute the valence
        deck_result = df['value'][trial_number]

        # compute valence
        valence = compute_valence(deck_result, W)

        # update expectancy of chosen deck
        expectancies[chosen_deck_index] = compute_expectancy(
            expectancies[chosen_deck_index], alpha, valence)

    return LL


def test_compute_log_likelihood(student_compute_log_likelihood, print_feedback=True):
    """ Returns True if The student implemented
    a compute_log_likelihood function that returns same values as the correct solution.

    Args
        compute_log_likelihood (method) :
            The student_compute_log_likelihood method defined in the notebook and edited by students.
        print_feedback (bool) :
            If True, this method prints feedback to the console.
    Returns
        correct (bool):
            True, if student_compute_log_likelihood returns the same values as compute_log_likelihood
    """

    correct = True

    wins = [5, 10, 10, 5]  # wins for Decks A, B, C, and D
    losses = [-5, -25, -25, -5]  # losses for Decks A, B, C, and D
    win_probability = 0.5  # chance of obtaining a win from a deck
    num_trials = 20

    for W in np.linspace(0, 1, 4):
        for alpha in np.linspace(0, 1, 4):
            for c in np.linspace(-1, 1, 4):

                wins = [5, 10, 10, 5]  # wins for Decks A, B, C, and D
                losses = [-5, -25, -25, -5]  # losses for Decks A, B, C, and D
                win_probability = 0.5  # chance of obtaining a win from a deck
                num_trials = 20
                random.seed(1)
                df = simulate_model(num_trials, wins, losses,
                                    win_probability, W, alpha, c)

                student_result = student_compute_log_likelihood(
                    df, W, alpha, c)
                correct_result = compute_log_likelihood(df, W, alpha, c)

                if np.round(student_result, decimals=8) != np.round(correct_result, decimals=8):
                    correct = False
                    break

        if not correct:
            break  # exit search

    if not print_feedback:
        return correct

    if correct:
        print("Your compute_log_likelihood function creates correct outputs!")
    else:
        print(
            "Your compute_log_likelihood function generates incorrect outputs. Check for mistakes.")
    return correct


def test_parameter_fit(df,
                       c_values,
                       W_values,
                       alpha_values,
                       student_LL,
                       student_c,
                       student_W,
                       student_alpha,
                       print_feedback=True):
    """ Returns True if computed the correct likelihood values for their data.

    Args
        df (pandas.DataFrame) :
            Student's data.
        c_values (list) :
            List of c values to test.
        W_values (list) :
            List of W values to test.
        alpha_values (list) :
            List of alpha values to test.
        student_LL (pandas.DataFrame) :
            Student's data.
        student_LL (float) :
            Student's log likelihood value
        student_c (float) :
            Student's fitted c value
        student_W (float) :
            Student's fitted W value
        student_alpha (float) :
            Student's fitted alpha value
        print_feedback (bool) :
            If True, this method prints feedback to the console.
    Returns
        correct (bool):
            True, if the student's likelihood and parameter values are correct.
    """

    correct = True

    # initialize log-likelihood and fitted parameters
    best_LL = -10000
    best_c = 0
    best_W = 0
    best_alpha = 0

    for c_idx, c in enumerate(c_values):
        for W_idx, W in enumerate(W_values):
            for alpha_idx, alpha in enumerate(alpha_values):

                LL = compute_log_likelihood(df, W, alpha, c)
                if LL > best_LL:
                    best_c = c
                    best_W = W
                    best_alpha = alpha
                    best_LL = LL

    if np.round(student_LL, decimals=8) != np.round(best_LL, decimals=8):
        correct = False
        print("Your log likelihood is incorrect.")
    if np.round(student_c, decimals=8) != np.round(best_c, decimals=8):
        correct = False
        print("Your c value is incorrect.")
    if np.round(student_W, decimals=8) != np.round(best_W, decimals=8):
        correct = False
        print("Your W value is incorrect.")
    if np.round(student_alpha, decimals=8) != np.round(best_alpha, decimals=8):
        correct = False
        print("Your alpha value is incorrect.")

    if not print_feedback:
        return correct

    if correct:
        print("Your compute_log_likelihood function creates correct outputs!")
    else:
        print(
            "Your compute_log_likelihood function generates incorrect outputs. Check for mistakes.")
    return correct


